//----------------------------------------------------------------------------
// Templated double linked list class
//----------------------------------------------------------------------------
#ifndef TEMPLATE_DOUBLE_LINK_LIST
#define TEMPLATE_DOUBLE_LINK_LIST

//----------------------------------------------------------------------------
#include "detl.h"
#include "tcontain.h"
#include "tqueue.h"
#include <assert.h>

//----------------------------------------------------------------------------
// Templated double link node for use with templated double link list
template <class T> class Tdouble_link_node
      {
      private:
            // The object and some links
            T node;
            Tdouble_link_node<T> *next, *prev;

            // Private constructors
            Tdouble_link_node (void)   : next(this),prev(this) { };
            Tdouble_link_node (T elem) : node(elem),next(this),prev(this) { };

      public:
            // The only thing we can do to a node is convert it to it's type
            inline operator T& () { return node; };

      // The Double linked list and derived classes can manipulate the internal nodes
      friend class Tdouble_link_list<T>;
      friend class Tqueue<T>;
      };

//----------------------------------------------------------------------------
// Templated double link list
template <class T> class Tdouble_link_list
      {
      private:
            // The head node, it's node is never used
            Tdouble_link_node<T> head;

            // Helper functions
            static void append (const T&, void*);
            void insert  (const T&, Tdouble_link_node<T>&);

      public:
            // Constructors
            Tdouble_link_list (void)          { };
            virtual ~Tdouble_link_list (void) { };

            const Tdouble_link_list<T>& operator=(const Tdouble_link_list<T>&);

            // Information functions
            inline bool is_empty (void) const  { if(head.next==&head) return true; else return false; };
                   int  number   (void) const;

            T* first_that (const Tcompare_function<T>&, void*) const;
            T* last_that  (const Tcompare_function<T>&, void*) const;

            // Indexing operations
            inline Tdouble_link_node<T>& first (void) { return *(head.next); };
            inline Tdouble_link_node<T>& last  (void) { return *(head.prev); };
            inline Tdouble_link_node<T>& next  (const Tdouble_link_node<T> &node) { return *(node.next); };
            inline Tdouble_link_node<T>& prev  (const Tdouble_link_node<T> &node) { return *(node.prev); };

            const T& operator [] (const int i) const;
            T& operator [] (const int i);

            // Manipulation functions
            inline void append  (const T& item) { insert(item,head);         };
                   void append  (const T& item, Tdouble_link_node<T>& node) { insert(item,*(node.next)); };
            inline void prepend (const T& item) { insert(item,*(head.next)); };
                   void remove  (Tdouble_link_node<T>&);
            inline void remove  (void)          { remove(*(head.prev)); };
            inline void clear   (void)          { while(is_empty()==false) remove(first()); };

            void for_each (const Titer_function<T>&, void*) const;
      };

//----------------------------------------------------------------------------
// Insert an item before another one
template <class T> void Tdouble_link_list<T>::insert (const T& item, Tdouble_link_node<T> &before)
   {
   Tdouble_link_node<T> *newNode = new Tdouble_link_node<T>(item);
   assert(newNode!=0);

   newNode->next = &before;
   newNode->prev = before.prev;
   before.prev->next = newNode;
   before.prev = newNode;
   }
//----------------------------------------------------------------------------
// Copy operation
template <class T> const Tdouble_link_list<T>& Tdouble_link_list<T>::operator=(const Tdouble_link_list<T>& other)
   {
   Titer_function<T> method(Tdouble_link_list::append);

   if(this!=&other)
      {
      if(is_empty()==false) clear();
      other.for_each(method,this);
      }
   return *this;
   }
//----------------------------------------------------------------------------
// Removes the given node
template <class T> void Tdouble_link_list<T>::remove (Tdouble_link_node<T>& node)
   {
   // We can't remove the head!
   assert(&node!=&head);

   node.prev->next = node.next;
   node.next->prev = node.prev;

   delete &node;
   }

//----------------------------------------------------------------------------
// Executes a given function with arguments on each item in the list
template <class T> void Tdouble_link_list<T>::for_each (const Titer_function<T>& func, void *args) const
   {
   // Get the first node in the list
   Tdouble_link_node<T> *node = head.next;

   // For each node in the list
   while(node!=&head)
      {
      // apply the given function to the node's item
      func(node->node,args);

      // get next node in the list
      node = node->next;
      }
   }
//----------------------------------------------------------------------------
// Return how many elements in the list
template <class T> int Tdouble_link_list<T>::number (void) const
   {
   Tdouble_link_node<T> *node = head.next;

   int i = 0;
   while(node!=&head)
      {
      i++;
      node = node->next;
      }
   return i;
   }

//----------------------------------------------------------------------------
// Index into the list with a constant
template <class T> const T& Tdouble_link_list<T>::operator [] (const int index) const
   {
   Tdouble_link_node<T> *node = head.next;

   int i = 0;
   while(i<index)
      {
      // Get the next node
      node = node->next;

      // Here we can skip over the head to make the indexes cyclic
      if(node!=&head) i++;
      }
   return *node;
   }
//----------------------------------------------------------------------------
// Index into the list
template <class T> T& Tdouble_link_list<T>::operator [] (const int index)
   {
   Tdouble_link_node<T> *node = head.next;

   int i = 0;
   while(i<index)
      {
      // Get the next node
      node = node->next;

      // Here we can skip over the head to make the indexes cyclic
      if(node!=&head) i++;
      }
   return *node;
   }
//----------------------------------------------------------------------------
// Returns the first Node in the list which the given function returns an int != 0, on the given Nodes
template <class T> T* Tdouble_link_list<T>::first_that (const Tcompare_function<T>& match, void *args) const
   {
   Tdouble_link_node<T> *node = head.next;

   while(node!=&head)
      {
      if(match(node->node,args)) return &(node->node);
      node = node->next;
      }
   return 0;
   }
//----------------------------------------------------------------------------
// Returns the last Node in the list which the given function returns an int != 0, on the given Nodes
template <class T> T* Tdouble_link_list<T>::last_that (const Tcompare_function<T>& match, void *args) const
   {
   Tdouble_link_node<T> *node = head.prev;

   while(node!=&head)
      {
      if(match(node->node,args)) return &(node->node);
      node = node->prev;
      }
   return 0;
   }
//----------------------------------------------------------------------------
// Static helper functions
//----------------------------------------------------------------------------
// Append an item to the list given
template <class T> void Tdouble_link_list<T>::append (const T& item, void *args)
   {
   Tdouble_link_list<T> *list = (Tdouble_link_list<T>*)args;
   list->append(item);
   }
//----------------------------------------------------------------------------
#endif

